const std = @import("std");
const dirs = @import("dirs.zig");
const json = @import("json");
const getty = @import("getty");

const Data = struct {
    versions: std.StringHashMap(std.SemanticVersion),
    in_use: ?[]u8 = null,
};

data: getty.de.Result(Data),

const Self = @This();

pub fn path(alloc: std.mem.Allocator) ![]const u8 {
    const home = try dirs.zupperHome(alloc);
    defer alloc.free(home);

    return try std.fs.path.join(alloc, &.{ home, "manifest.json" });
}

pub fn deinit(self: *Self) void {
    self.data.deinit();

    self.* = undefined;
}

pub fn load(alloc: std.mem.Allocator) !Self {
    const mpath = try path(alloc);
    defer alloc.free(mpath);

    const infile = std.fs.cwd().openFile(mpath, .{}) catch |e| blk: {
        if (e != error.FileNotFound)
            return e;

        std.log.warn("manifest not found! creating new one.", .{});
        break :blk null;
    };

    if (infile) |f| {
        defer f.close();

        // no support for streaming deserialization :(
        const data = try f.readToEndAlloc(alloc, std.math.maxInt(usize));
        defer alloc.free(data);

        const deser = json.fromSlice(alloc, Data, data) catch |e| {
            std.log.err(
                \\failed to deserialize manifest! someone messed with my files! (how dare you!): {s}
                \\try deleting '{s}'
            ,
                .{ @errorName(e), mpath },
            );
            return error.Explained;
        };

        return .{ .data = deser };
    } else {
        var arena = try alloc.create(std.heap.ArenaAllocator);
        arena.* = std.heap.ArenaAllocator.init(alloc);
        return .{ .data = .{
            .value = .{
                .versions = std.StringHashMap(std.SemanticVersion).init(arena.allocator()),
            },
            .arena = arena,
        } };
    }
}

pub fn setVersion(self: *Self, k: []const u8, v: ?[]const u8) !void {
    const k_d = try self.data.arena.allocator().dupe(u8, k);

    const gopr = try self.data.value.versions.getOrPut(k_d);

    // this is the actual version. this is used in the manifest
    // to keep the version of master up to date.
    var semver = std.SemanticVersion.parse(v orelse k) catch |e| {
        std.log.err("failed to parse zig semver: {s}", .{@errorName(e)});
        return error.Explained;
    };

    // we move the strings in the semver to the manifest allocator so getty can free them
    if (semver.pre) |x| semver.pre = try self.data.arena.allocator().dupe(u8, x);
    if (semver.build) |x| semver.build = try self.data.arena.allocator().dupe(u8, x);

    gopr.value_ptr.* = semver;
}

pub fn removeVersion(self: *Self, version: []const u8) void {
    _ = self.data.value.versions.remove(version);
}

pub fn setInUse(self: *Self, in_use: ?[]const u8) !void {
    if (in_use) |v| {
        const dup = try self.data.arena.allocator().dupe(u8, v);
        self.data.value.in_use = dup;
    }
}

pub fn save(self: *Self) !void {
    const mpath = try path(self.data.arena.child_allocator);
    defer self.data.arena.child_allocator.free(mpath);

    var outfile = try std.fs.cwd().createFile(mpath, .{});
    defer outfile.close();

    try json.toWriter(self.data.arena.child_allocator, self.data.value, outfile.writer());
}
