const std = @import("std");
const c = @import("../ffi.zig").c;
const dirs = @import("../dirs.zig");
const net = @import("../net.zig");
const Args = @import("../args.zig").Install;
const Manifest = @import("../Manifest.zig");

pub fn run(
    alloc: std.mem.Allocator,
    manifest: *Manifest,
    args: Args,
    positionals: []const [:0]const u8,
) !void {
    if (positionals.len < 1) {
        std.log.err(
            "expected at least 1 positional argument to command 'install', found {}",
            .{positionals.len},
        );
        return error.Explained;
    }

    if (args.use and positionals.len != 1) {
        std.log.err("expected exactly 1 argument when called with the '--use' flag", .{});
        return error.Explained;
    }

    const home = try dirs.zupperHome(alloc);
    defer alloc.free(home);

    for (positionals) |version| {
        if (try dirs.isVersionInstalled(alloc, home, version)) {
            std.log.err(
                "version '{s}' is already installed! try using the update subcommand to update it.",
                .{version},
            );
            return error.Explained;
        }
    }

    var http = std.http.Client{ .allocator = alloc };
    defer http.deinit();

    var version_index = try net.getVersionIndex(&http);
    defer version_index.deinit();

    for (positionals) |version_arg| {
        var exactv_buf: [1024]u8 = undefined;
        const version = version_index.value.get(version_arg) orelse exactv: {
            std.log.warn("no such version in index! trying exact version...", .{});

            const exactv_url = try std.fmt.bufPrint(
                &exactv_buf,
                "https://ziglang.org/builds/zig-{s}-{s}-{s}.tar.xz",
                .{
                    @tagName(@import("builtin").os.tag),
                    @tagName(@import("builtin").target.cpu.arch),
                    version_arg,
                },
            );

            std.log.info("exact version URL: '{s}'", .{exactv_url});

            var exactv_result = http.fetch(alloc, .{
                .location = .{ .url = exactv_url },
                .response_strategy = .none,
                .method = .HEAD,
            }) catch |e| {
                std.log.err("checking exact version: {}", .{e});
                return error.Explained;
            };
            defer exactv_result.deinit();

            if (exactv_result.status.class() != .success) {
                std.log.info("no such exact version available ({} {s})! available versions: ", .{
                    @intFromEnum(exactv_result.status),
                    exactv_result.status.phrase() orelse "<unknown>",
                });
                var kiter = version_index.value.keyIterator();
                while (kiter.next()) |v| {
                    std.log.info("\t{s}", .{v.*});
                }
                return error.Explained;
            }

            break :exactv net.Version{ .file = .{ .tarball = exactv_url } };
        };

        const version_file = version.file orelse {
            std.log.err("the version '{s}' has no file available for your system!", .{version_arg});
            return error.Explained;
        };

        const install_path = try std.fs.path.join(alloc, &.{ home, "installs", version_arg });
        defer alloc.free(install_path);

        try net.downloadVersionTo(&http, version_file.tarball, install_path);
        try manifest.setVersion(version_arg, version.version);
    }

    if (args.use) {
        try dirs.useVersion(alloc, manifest, home, positionals[0]);
    }
}
