const std = @import("std");
const getty = @import("getty");
const json = @import("json");
const c = @import("../ffi.zig").c;
const dirs = @import("../dirs.zig");
const net = @import("../net.zig");
const Args = @import("../args.zig").Update;
const Manifest = @import("../Manifest.zig");

pub fn run(
    alloc: std.mem.Allocator,
    manifest: *Manifest,
    args: Args,
    positionals: []const [:0]const u8,
) !void {
    _ = args;

    const home = try dirs.zupperHome(alloc);
    defer alloc.free(home);

    var http = std.http.Client{ .allocator = alloc };
    defer http.deinit();

    var version_index = try net.getVersionIndex(&http);
    defer version_index.deinit();

    var updated_any = false;

    var manifest_version_iter = manifest.data.value.versions.iterator();
    while (manifest_version_iter.next()) |version| {
        // if the user provided arguments, only update those versions
        if (positionals.len != 0) {
            const should_update = blk: {
                for (positionals) |version_arg| {
                    if (std.mem.eql(u8, version_arg, version.key_ptr.*))
                        break :blk true;
                }
                break :blk false;
            };

            if (!should_update)
                continue;
        }

        const idx_version = version_index.value.get(version.key_ptr.*) orelse {
            std.log.warn(
                "the installed version '{s}' is not in the version index",
                .{version.key_ptr.*},
            );
            continue;
        };

        const version_file = idx_version.file orelse {
            std.log.err(
                "the version '{s}' has no file available for your system! skipping!",
                .{version.key_ptr.*},
            );
            return error.Explained;
        };

        var idx_semver = std.SemanticVersion.parse(
            // if the version index doesn't provide a version, we skip it,
            // as it's not a version that should auto-update
            idx_version.version orelse continue,
        ) catch |e| {
            std.log.err("failed to parse semantic version in index: {s}", .{@errorName(e)});
            return error.Explained;
        };

        updated_any = true;

        switch (idx_semver.order(version.value_ptr.*)) {
            .lt => {
                std.log.warn(
                    \\installed version '{s}' is newer than version in index!
                    \\thank you for taking zupper on your time travelling adventures!
                ,
                    .{version.key_ptr.*},
                );
                continue;
            },
            .eq => {
                std.log.info("installed version '{s}' is up to date", .{version.key_ptr.*});
                continue;
            },
            .gt => {},
        }

        std.log.info("deleting old version for '{s}'", .{version.key_ptr.*});

        const install_path = try std.fs.path.join(alloc, &.{ home, "installs", version.key_ptr.* });
        defer alloc.free(install_path);

        try std.fs.cwd().deleteTree(install_path);

        try net.downloadVersionTo(&http, version_file.tarball, install_path);

        // we don't use manifest.setVersion here because that'd be kinda dodgy as we're
        // iterating the map here
        if (idx_semver.pre) |x| idx_semver.pre = try manifest.data.arena.allocator().dupe(u8, x);
        if (idx_semver.build) |x| idx_semver.build = try manifest.data.arena.allocator().dupe(u8, x);

        version.value_ptr.* = idx_semver;
    }

    if (!updated_any) {
        std.log.err("no versions to update!", .{});
        return error.Explained;
    }
}
