const std = @import("std");
const kf = @import("known-folders");
const Manifest = @import("Manifest.zig");

pub fn zupperHome(alloc: std.mem.Allocator) ![]const u8 {
    const datapath = (try kf.getPath(alloc, .data)) orelse return error.GetDataFolder;
    defer alloc.free(datapath);

    const folder = try std.fs.path.join(alloc, &.{ datapath, "zupper" });

    // create the folder so subcommands can work with it.
    try std.fs.cwd().makePath(folder);

    return folder;
}

pub fn isVersionInstalled(alloc: std.mem.Allocator, home: []const u8, version: []const u8) !bool {
    const path = try std.fs.path.join(alloc, &.{ home, "installs", version });
    defer alloc.free(path);

    std.fs.cwd().access(path, .{}) catch return false;
    return true;
}

pub fn binInstallationPath(alloc: std.mem.Allocator) !?[]u8 {
    switch (@import("builtin").os.tag) {
        .linux => {
            const user_home = (try kf.getPath(alloc, .home)) orelse return error.GetHomeDir;
            defer alloc.free(user_home);
            return try std.fs.path.join(alloc, &.{ user_home, ".local", "bin" });
        },
        .windows => {
            return try std.fs.selfExeDirPathAlloc(alloc);
        },
        .macos => {
            const user_home = (try kf.getPath(alloc, .home)) orelse return error.GetHomeDir;
            defer alloc.free(user_home);
            // not sure if this is the correct path. mac got kind of a windowsy situation too
            return try std.fs.path.join(alloc, &.{ user_home, "Applications" });
        },
        else => {
            return null;
        },
    }
}

pub fn useVersion(
    alloc: std.mem.Allocator,
    manifest: *Manifest,
    home: []const u8,
    version: []const u8,
) !void {
    if (manifest.data.value.in_use != null and std.mem.eql(u8, manifest.data.value.in_use.?, version)) {
        std.log.err("version '{s}' is already in use!", .{version});
        return error.Explained;
    }

    if (!try isVersionInstalled(alloc, home, version)) {
        std.log.err(
            "version '{s}' isn't installed! have you tried installing it first?",
            .{version},
        );
        return error.Explained;
    }

    const zig_exe_path = try std.fs.path.join(alloc, &.{
        home,
        "installs",
        version,
        comptime "zig" ++ @import("builtin").target.exeFileExt(),
    });
    defer alloc.free(zig_exe_path);

    std.fs.cwd().access(zig_exe_path, .{}) catch |e| {
        std.log.err(
            \\failed to access zig executable @ '{s}': {s}
            \\it looks like the version '{s}' might be broken. try uninstalling it.
        ,
            .{ zig_exe_path, @errorName(e), version },
        );
        return error.Explained;
    };

    if (@import("builtin").os.tag == .windows) {
        std.log.warn(
            \\installing on windows! windows is so bad, it can't get it's conventions
            \\for locally installed programs right. i'll use my own install directory...
        ,
            .{},
        );
    }

    const target_binary_directory = try binInstallationPath(alloc) orelse {
        std.log.err(
            "using a version isn't supported on {s}. please install the zig binary @ '{s}' manually",
            .{ @tagName(@import("builtin").os.tag), zig_exe_path },
        );

        return;
    };
    defer alloc.free(target_binary_directory);

    const target_binary = try std.fs.path.join(alloc, &.{
        target_binary_directory,
        comptime "zig" ++ @import("builtin").target.exeFileExt(),
    });
    defer alloc.free(target_binary);

    std.log.info("installing version '{s}' to '{s}'", .{ version, target_binary });

    std.fs.cwd().deleteFile(target_binary) catch |e| {
        if (e != error.FileNotFound) {
            return e;
        }
    };
    try std.fs.cwd().symLink(zig_exe_path, target_binary, .{});
    try manifest.setInUse(version);
}

pub fn unuseCurrentVersion(
    alloc: std.mem.Allocator,
    manifest: *Manifest,
) !void {
    const install_dir = try binInstallationPath(alloc) orelse return;
    defer alloc.free(install_dir);

    const target_binary = try std.fs.path.join(alloc, &.{
        install_dir,
        comptime "zig" ++ @import("builtin").target.exeFileExt(),
    });
    defer alloc.free(target_binary);

    std.log.info("unusing zig installation @ {s}", .{target_binary});
    try std.fs.cwd().deleteFile(target_binary);

    manifest.data.value.in_use = null;
}
