const std = @import("std");
const c = @import("ffi.zig").c;
const ansiterm = @import("ansi-term");
const getty = @import("getty");
const json = @import("json");
const Manifest = @import("Manifest.zig");

pub const VersionFile = struct {
    tarball: []const u8,

    pub const @"getty.db" = struct {
        pub const attributes = .{
            .Container = .{ .ignore_unknown_fields = true },
        };
    };
};

pub const Version = struct {
    /// this field only exists for master
    version: ?[]const u8 = null,

    file: ?VersionFile = null,

    pub const @"getty.db" = struct {
        pub const attributes = .{
            .Container = .{ .ignore_unknown_fields = true },
            .file = .{
                // We deserialize only the file information for the current architecture.
                // (Ain't that cool? :P)
                .rename = @tagName(@import("builtin").target.cpu.arch) ++
                    "-" ++
                    @tagName(@import("builtin").os.tag),
            },
        };
    };
};

pub const VersionMap = std.StringHashMap(Version);

pub fn getVersionIndex(http: *std.http.Client) !getty.de.Result(VersionMap) {
    const alloc = http.allocator;
    std.log.info("getting version index", .{});

    var req = try http.open(
        .GET,
        comptime std.Uri.parse("https://ziglang.org/download/index.json") catch unreachable,
        std.http.Headers.init(alloc), // freed by req.deinit
        .{},
    );
    defer req.deinit();

    try req.send(.{});
    try req.wait();

    return try json.fromReader(alloc, VersionMap, req.reader());
}

/// concurrently downloads and unpacks a version to the given path
/// `alloc` must be thread-safe!
pub fn downloadVersionTo(
    http: *std.http.Client,
    url_str: []const u8,
    target: []const u8,
) !void {
    const alloc = http.allocator;

    const url = try std.Uri.parse(url_str);

    std.log.info(
        \\downloading version:
        \\ /< {s}
        \\ \> {s}
    ,
        .{ url_str, target },
    );

    var req = try http.open(.GET, url, std.http.Headers.init(alloc), .{});
    defer req.deinit();

    try req.send(.{});
    try req.wait();

    const DLState = struct {
        req: *std.http.Client.Request,
        target: []const u8,
        dltotal: u64,
        dlnow: u64 = 0,
        current_fname_buf: [512]u8 = undefined,
        current_fname_len: usize = 0,
        last_status: i64 = 0,
        read_buf: [1024 * 8]u8 = undefined,

        const DLState = @This();

        fn run(self: *DLState) !void {
            // TODO: figure out how to use std.tar and yoink libarchive
            const archive = c.archive_read_new();
            defer _ = c.archive_read_free(archive);

            try handleArchiveErr(c.archive_read_support_filter_all(archive), archive);
            try handleArchiveErr(c.archive_read_support_format_all(archive), archive);

            try handleArchiveErr(c.archive_read_open(
                archive,
                self,
                null,
                &archiveReadCB,
                null,
            ), archive);

            var buf: [512]u8 = undefined;
            var entry: ?*c.archive_entry = null;

            while (c.archive_read_next_header(archive, &entry) == c.ARCHIVE_OK) {
                const entry_path = std.mem.span(c.archive_entry_pathname(entry));
                var trimmed_entry_path = entry_path;

                // no need to include the archive name in the path again
                if (std.mem.startsWith(u8, entry_path, "zig-")) {
                    const first_slash = std.mem.indexOfScalar(u8, trimmed_entry_path, '/');
                    if (first_slash) |fs| {
                        if (fs == entry_path.len - 1) {
                            // If the only slash is at the end, this is the archive root entry.
                            // We don't need this.
                            continue;
                        }
                        trimmed_entry_path = entry_path[(fs + 1)..];
                    }
                }

                // TODO: path traversal?
                const fpath = try std.fs.path.join(self.req.client.allocator, &.{
                    self.target,
                    trimmed_entry_path,
                });
                defer self.req.client.allocator.free(fpath);

                // entries that end in a slash are directories
                if (fpath[fpath.len - 1] == '/') {
                    try std.fs.cwd().makePath(fpath);
                    continue;
                }

                const entry_basename = std.fs.path.basename(trimmed_entry_path);
                if (entry_basename.len <= self.current_fname_buf.len) {
                    @memcpy(self.current_fname_buf[0..entry_basename.len], entry_basename);
                    self.current_fname_len = entry_basename.len;
                }

                const file = try std.fs.cwd().createFile(
                    fpath,
                    .{ .mode = c.archive_entry_mode(entry) },
                );
                defer file.close();

                while (true) {
                    const status = c.archive_read_data(archive, &buf, buf.len);
                    if (status < 0) {
                        try handleArchiveErr(status, archive);
                    } else if (status == 0) {
                        break;
                    } else {
                        try file.writeAll(buf[0..@intCast(status)]);
                    }
                }
            }
        }

        fn archiveReadCB(
            archive: ?*c.archive,
            udata: ?*anyopaque,
            buf: ?*?*const anyopaque,
        ) callconv(.C) isize {
            const self: *DLState = @ptrCast(@alignCast(udata));
            _ = archive;

            // status update
            status: {
                const now = std.time.milliTimestamp();

                // don't do anything if the last update wasn't at least 100ms ago
                if (now - self.last_status < 100)
                    break :status;

                self.last_status = now;

                self.writeStatus() catch |e| {
                    std.log.err("logging: {}", .{e});
                    return c.ARCHIVE_FATAL;
                };
            }

            buf.?.* = &self.read_buf;
            const read = self.req.read(&self.read_buf) catch |e| {
                std.log.err("reading data from HTTP: {}", .{e});
                return c.ARCHIVE_FATAL;
            };
            self.dlnow += read;
            return @intCast(read);
        }

        fn writeStatus(self: *DLState) !void {
            if (self.current_fname_len == 0)
                return;

            const percent = if (self.dltotal != 0) @divTrunc(self.dlnow * 100, self.dltotal) else 0;

            var stdout_buf = std.io.bufferedWriter(std.io.getStdOut().writer());

            const stdout = stdout_buf.writer();

            try ansiterm.clear.clearCurrentLine(stdout);
            try ansiterm.cursor.setCursorColumn(stdout, 0);

            try ansiterm.format.updateStyle(
                stdout,
                .{ .foreground = .Green, .font_style = .{ .bold = true } },
                null,
            );

            try stdout.print("[{d:3}%] ", .{percent});

            try ansiterm.format.updateStyle(
                stdout,
                .{ .foreground = .Blue },
                .{ .foreground = .Green, .font_style = .{ .bold = true } },
            );
            try stdout.writeAll(self.current_fname_buf[0..self.current_fname_len]);

            try ansiterm.format.resetStyle(stdout);

            try stdout_buf.flush();
        }
    };

    var state = DLState{
        .req = &req,
        .target = target,
        .dltotal = req.response.content_length orelse 0,
    };

    try std.fs.cwd().makePath(target);

    try state.run();

    const stdout = std.io.getStdOut().writer();
    try ansiterm.format.updateStyle(
        stdout,
        .{ .foreground = .Green, .font_style = .{ .bold = true } },
        null,
    );
    try ansiterm.cursor.setCursorColumn(stdout, 0);
    try ansiterm.clear.clearCurrentLine(stdout);
    try stdout.writeAll("[100%]\n");
    try ansiterm.format.resetStyle(stdout);
}

fn handleArchiveErr(err: c_long, archive: ?*c.archive) !void {
    if (err != c.ARCHIVE_OK) {
        if (archive) |ar| {
            if (c.archive_error_string(ar)) |err_s|
                std.log.err("archive: {s}", .{err_s});
        }
        return error.Explained;
    }
}
